/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CombineMap;

DECLARE_NODE(CombineMap)
/*Node
 *  
 * @name CombineMap
 * @category HARK:MISC
 * @description Combine two (or more -> not yet) maps.
 *
 * @input_name INPUT1
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Vector<float>, Vector<complex<float> >, Matrix<float> or Matrix<complex<float> >.
 *
 * @input_name INPUT2
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Vector<float>, Vector<complex<float> >, Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>	
 * @output_description  The key is source ID, and the value is Vector<float>, Vector<complex<float> >, Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name INPUT_PRIORITY
 * @parameter_type string
 * @parameter_value TOP
 * @parameter_list TOP:BOTTOM
 * @parameter_description When the key is duplicated, "TOP" uses first value,"BOTTOM" uses latest value in inputs. [default: TOP]
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int PRIORITY_TOP = 0;
static const int PRIORITY_BOTTOM = 1;

static const int INPUT_VECTOR_FLOAT = 0;
static const int INPUT_VECTOR_COMPLEX = 1;
static const int INPUT_MATRIX_FLOAT = 2;
static const int INPUT_MATRIX_COMPLEX = 3;

class CombineMap : public BufferedNode {
	int input1ID;
	int input2ID;
	int outputID;
	int priority;
	bool debug;

public:
	CombineMap(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) { 
		input1ID = addInput("INPUT1");
		input2ID = addInput("INPUT2");
		outputID = addOutput("OUTPUT");
		string method_name = object_cast<String>(parameters.get("INPUT_PRIORITY"));
		if (method_name == "TOP") {
			priority = PRIORITY_TOP;
		}
		else if (method_name == "BOTTOM") {
			priority = PRIORITY_BOTTOM;
		}
		else {
			cerr << "******* CombineMap: Input error" << endl;
			throw new NodeException(NULL, "CombineMap: INPUT_PRIORITY should be \"TOP\" or \"BOTTOM\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));

		inOrder = true;
	}
	
	~CombineMap() {
	}

	void calculate(int output_id, int count, Buffer &out) {
		if (debug == true) cout << "******* MapToMap debug point-Input check (count=" << count << ")" << endl;

		RCPtr<Map<int, ObjectRef> > input1(getInput(input1ID, count));
		RCPtr<Map<int, ObjectRef> > input2(getInput(input2ID, count));

		int typeOR;
		bool typeMatch = false;
		Map<int, ObjectRef>::const_iterator it;
		int inputId;
		ObjectRef inputOR;
		int sizeOR1, rowsOR1, colsOR1, sizeOR2, rowsOR2, colsOR2;
		bool sizeMatch = false;

		if ((*input1).empty() == true) {
			if (debug == true) cout << "Input1 map empty" << endl;
		}
		else {
			it = (*input1).begin();
			inputId = it->first;
			inputOR = it->second;
			if (debug == true) {
				cout << "Input1 begins from " << inputId << endl;
				cout << "Input1 - Number of channels: " << (*input1).size() << endl;
				cout << "Input1 - Type of ObjectRef at first channel: ";
			}
			if (typeid(*inputOR) == typeid(Vector<float>)) {
				typeOR = INPUT_VECTOR_FLOAT;
				sizeOR1 = (object_cast<Vector<float> >(inputOR)).size();
		    	if (debug == true) {
					cout << "Vector<float>" << endl;
		    		cout << "Input1 - Size of ObjectRef st first channel: " << sizeOR1 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Vector<complex<float> >)) {
				typeOR = INPUT_VECTOR_COMPLEX;
				sizeOR1 = (object_cast<Vector<complex<float> > >(inputOR)).size();
				if (debug == true) {
					cout << "Vector<complex<float> >" << endl;
		    		cout << "Input1 - Size of ObjectRef at first channel: " << sizeOR1 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Matrix<float>)) {
				typeOR = INPUT_MATRIX_FLOAT;
				rowsOR1 = (object_cast<Matrix<float> >(inputOR)).nrows();
				colsOR1 = (object_cast<Matrix<float> >(inputOR)).ncols();
				if (debug == true) {
					cout << "Matrix<float>" << endl;
		    		cout << "Input1 - Size of ObjectRef at first channel: " << rowsOR1 << " x " << colsOR1 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Matrix<complex<float> >)) {
				typeOR = INPUT_MATRIX_COMPLEX;
				rowsOR1 = (object_cast<Matrix<complex<float> > >(inputOR)).nrows();
				colsOR1 = (object_cast<Matrix<complex<float> > >(inputOR)).ncols();
				if (debug == true) {
					cout << "Matrix<complex<float> >" << endl;
		    		cout << "Input1 - Size of ObjectRef at first channel: " << rowsOR1 << " x " << colsOR1 << endl;
				}
			}
			else {
				if (debug == true) cout << "unidentified." << endl;
				cerr << "******* CombineMap: Input error" << endl;
				throw new NodeException(this, "Type of ObjectRef in Input map is not supported.", __FILE__, __LINE__);
			}
		}
		if ((*input2).empty() == true) {
			if (debug == true) cout << "Input2 map empty" << endl;
		}
		else {
			it = (*input2).begin();
			inputId = it->first;
			inputOR = it->second;
			if (debug == true) {
				cout << "Input2 begins from " << inputId << endl;
				cout << "Input2 - Number of channels: " << (*input2).size() << endl;
				cout << "Input2 - Type of ObjectRef at first channel: ";
			}
			if (typeid(*inputOR) == typeid(Vector<float>)) {
				if (typeOR == INPUT_VECTOR_FLOAT) typeMatch = true;
				sizeOR2 = (object_cast<Vector<float> >(inputOR)).size();
				if (sizeOR1 == sizeOR2) sizeMatch = true;
		    	if (debug == true) {
					cout << "Vector<float>" << endl;
		    		cout << "Input2 - Size of ObjectRef at first channel: " << sizeOR2 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Vector<complex<float> >)) {
				if (typeOR == INPUT_VECTOR_COMPLEX) typeMatch = true;
				sizeOR2 = (object_cast<Vector<complex<float> > >(inputOR)).size();
				if (sizeOR1 == sizeOR2) sizeMatch = true;
		    	if (debug == true) {
					cout << "Vector<complex<float> >" << endl;
		    		cout << "Input2 - Size of ObjectRef at first channel: " << sizeOR2 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Matrix<float>)) {
				if (typeOR == INPUT_MATRIX_FLOAT) typeMatch = true;
				rowsOR2 = (object_cast<Matrix<float> >(inputOR)).nrows();
				colsOR2 = (object_cast<Matrix<float> >(inputOR)).ncols();
				if (rowsOR1 == rowsOR2 && colsOR1 == colsOR2) sizeMatch = true;
		    	if (debug == true) {
					cout << "Matrix<float>" << endl;
		    		cout << "Input2 - Size of ObjectRef at first channel: " << rowsOR2 << " x " << colsOR2 << endl;
				}
			}
			else if (typeid(*inputOR) == typeid(Matrix<complex<float> >)) {
				if (typeOR == INPUT_MATRIX_COMPLEX) typeMatch = true;
				rowsOR2 = (object_cast<Matrix<complex<float> > >(inputOR)).nrows();
				colsOR2 = (object_cast<Matrix<complex<float> > >(inputOR)).ncols();
				if (rowsOR2 == rowsOR2 && colsOR1 == colsOR2) sizeMatch = true;
		    	if (debug == true) {
					cout << "Matrix<complex<float> >" << endl;
		    		cout << "Input2 - Size of ObjectRef at first channel: " << rowsOR1 << " x " << colsOR2 << endl;
				}
			}
			else {
				if (debug == true) cout << "nidentified." << endl;
				cerr << "******* CombineMap: Input error" << endl;
				throw new NodeException(this, "Type of ObjectRef in Input map is not supported.", __FILE__, __LINE__);
			}
			if (not typeMatch) {
				cerr << "******* CombineMap: Input error" << endl;
				throw new NodeException(this, "Types of ObjectRef in Input maps are mismatch.", __FILE__, __LINE__);
			}
			if (not sizeMatch) {
				cerr << "******* CombineMap: Input error" << endl;
				throw new NodeException(this, "Sizes of ObjectRef in Input maps are mismatch.", __FILE__, __LINE__);
			}
		}

		ObjectRef orOutput = ObjectRef(new Map<int, ObjectRef>);
		Map<int, ObjectRef> &mpOutput = object_cast<Map<int, ObjectRef> >(orOutput);
		out[count] = orOutput;

		for (it = (*input1).begin(); it != (*input1).end(); ++it) {
			mpOutput.insert(make_pair(it->first, it->second));
		}
		for (it = (*input2).begin(); it != (*input2).end(); ++it) {
			if (mpOutput.count(it->first) > 0 && priority == PRIORITY_BOTTOM) {
				mpOutput.erase(it->first);
			}
			mpOutput.insert(make_pair(it->first, it->second));
		}
		
		if (debug == true) cout << "******* CombineMap debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(CombineMap);
};
