/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <fstream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include <string>
#include <sstream>

//#include <boost/tokenizer.hpp>

using namespace std;
using namespace FD;
//using namespace boost;

class LoadMatrixFrames;

DECLARE_NODE(LoadMatrixFrames);
/*Node
 *
 * @name LoadMatrixFrames
 * @category HARK:MISC
 * @description Load matrix data of frames from file
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Data Stream
 *
 * @output_name NOT_EOF
 * @output_type bool
 * @output_description True if the current frame does not reach the End-of-File.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The filename to load.
 *
 * @parameter_name INPUT_TYPE
 * @parameter_type string
 * @parameter_list TEXT_float:RAW_float:TEXT_complex_float:RAW_complex_float
 * @parameter_value TEXT_float
 * @parameter_description Input file format type [default: TEXT_float]
 *
 * @parameter_name ROW_SIZE
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description ROW size of Matrix
 *
 * @parameter_name COL_SIZE
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description COL size of Matrix
 *
 END*/

class LoadMatrixFrames : public BufferedNode {
     #define ENABLE_LOG 0
    int outputID;
    int noteofID;

    typedef enum {TEXT_float, RAW_float, TEXT_complex_float, RAW_complex_float} InputMode;

    InputMode mode;
    string inputType;
    string filename;
    int row;
    int col;
    ifstream fin;
    int length;	
    int line;
    bool isObject;
    ofstream file;

    public:
    LoadMatrixFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT_float), row(8), col(512), length(0), line(0), isObject(true)
    {
        outputID = addOutput("OUTPUT");
        noteofID = addOutput("NOT_EOF");
        inputType = object_cast<String> (parameters.get("INPUT_TYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));
        row = dereference_cast<int>(parameters.get("ROW_SIZE"));
        col = dereference_cast<int>(parameters.get("COL_SIZE"));
        
        fin.open(filename.c_str(), ios::binary);
        if(fin.fail())
        {
            throw new NodeException(NULL, string("Can't open '")+filename+string("'."), __FILE__, __LINE__);
        }

	    fin.seekg (0, fin.end);
        length = fin.tellg();
        fin.seekg (0, fin.beg);
	    line = ((length /col) / row) / sizeof(float);
        cout << "ROW[" << row << "] COL[" << col << "] total[" << (row * col * 2) << "]  length[" << length << "]" << endl;

        if (inputType.compare("RAW_float") == 0) {
            mode = RAW_float;
        }
        else if(inputType.compare("TEXT_complex_float") == 0) {
            mode = TEXT_complex_float;
	        line = line/2;
        }
        else if(inputType.compare("RAW_complex_float") == 0) {
            mode = RAW_complex_float;
	        line = line/2;
        }
        else if(inputType.compare("TEXT_float") == 0) {
            mode = TEXT_float;
        }

        // This file is for debugging.  Will remove lines that does operation for this file once deployed.
        if(ENABLE_LOG) file.open("load.txt");
    }

    ~LoadMatrixFrames()
    {
        fin.close();
    }

    virtual void initialize()
    {
        outputs[outputID].lookAhead =
        outputs[noteofID].lookAhead = 1 + max(outputs[outputID].lookAhead,outputs[noteofID].lookAhead);

        outputs[outputID].lookBack =
        outputs[noteofID].lookBack =  1 + max(outputs[outputID].lookBack,outputs[noteofID].lookBack);

        this->BufferedNode::initialize();
    }

    bool transferTextToContainer(int count, vector<string>& tokens) {
        string buffer;
        ifstream::pos_type cp;
        cp = fin.tellg();
        getline(fin, buffer);

        if(fin.eof()) {
         (*(outputs[noteofID].buffer))[count] = FalseObject; }

        if(buffer.empty())
            return false;
        //typedef boost::tokenizer<boost::char_separator<char> > tokenizer_space;
        //boost::char_separator<char> sep(" ");
        //tokenizer_space tokenizer(buffer, sep);
        char sep = ' ';
        stringstream bufferStream(buffer);
        string token;
        //for (tokenizer_space::iterator it = tokenizer.begin(); it != tokenizer.end(); ++it) {
        //    tokens.push_back(*it);
        while (getline(bufferStream, token, sep)) {
            tokens.push_back(move(token));
        }

        return true;
    }

    void calculate(int output_id, int count, Buffer &out) {
        (*(outputs[noteofID].buffer))[count] = ((fin.eof() || (isObject == false)) ? FalseObject : TrueObject);

        switch(mode)
        {
            case TEXT_float:
            {
                RCPtr <Matrix<float> > paramout(new Matrix<float>(row, col));
                (*(outputs[outputID].buffer))[count] = paramout;

                vector<string> tokens;
                if(transferTextToContainer(count, tokens)) {
                    if (tokens.size() != (row * col)) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL/ROW values if set correctly.", __FILE__, __LINE__);
                    }

                    for (unsigned int cnt1 = 0; cnt1 < row; cnt1++) {
                        for (unsigned int cnt2 = 0; cnt2 < col; cnt2++) {
                            char *endptr;
                            float val = strtof(tokens[cnt2 + cnt1 * col].c_str(), &endptr);
                            if (endptr == "\0")
                                cout << "The string that was not converted is " << endptr << "." << endl;
                            (*paramout)(cnt1, cnt2) = val;
                            if(ENABLE_LOG) file << (*paramout)(cnt1, cnt2) << " ";
                        }
                    }
                        if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case RAW_float:
            {
                RCPtr <Matrix<float> > paramout(new Matrix<float>(row, col));
                (*(outputs[outputID].buffer))[count] = paramout;

                std::vector<float> buf(row * col);
                fin.read((char *)&buf[0], buf.size() * sizeof(float));
                int actual_size = fin.gcount();
                if(!(actual_size)) {
                    isObject = false;
                } else {
                    if(actual_size != (row * col * sizeof(float))) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL/ROW values if set correctly.", __FILE__, __LINE__);
                    }
                    for (unsigned int cnt1 = 0; cnt1 < row; cnt1++) {
                        for (unsigned int cnt2 = 0; cnt2 < col; cnt2++) {
                            float val = buf[cnt2 + cnt1 * col];
                            (*paramout)(cnt1, cnt2) = val;
                            if(ENABLE_LOG) file << val << " ";
                        }
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case TEXT_complex_float:
            {
                RCPtr <Matrix<complex<float> > > paramout(new Matrix<complex<float> >(row, col));
                (*(outputs[outputID].buffer))[count] = paramout;

                vector<string> tokens;
                if(transferTextToContainer(count, tokens)) {
                    if (tokens.size() != (row * col * 2)) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL/ROW values if set correctly.", __FILE__, __LINE__);
                    }
 
                    for (unsigned int cnt1 = 0; cnt1 < row; cnt1++) {
                        for (unsigned int cnt2 = 0; cnt2 < col; cnt2++) {
                            char *endptr1, *endptr2;
                            string str_real = tokens[cnt2 * 2 + cnt1 * col * 2 + 0];
                            string str_imag = tokens[cnt2 * 2 + cnt1 * col * 2 + 1];
                            float tmp_r = strtof(str_real.c_str(), &endptr1);
                            if (endptr1 == "\0")
                                cout << "The string that was not converted is " << endptr1 << "." << endl;
                            float tmp_i = strtof(str_imag.c_str(), &endptr2);
                            if (endptr2 == "\0")
                            cout << "The string that was not converted is " << endptr2 << "." << endl;
                            (*paramout)(cnt1, cnt2) = complex<float>(tmp_r, tmp_i);
                            if(ENABLE_LOG) file << (*paramout)(cnt1, cnt2) << " ";
                        }
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case RAW_complex_float:
            {
                RCPtr <Matrix<complex<float> > > paramout(new Matrix<complex<float> >(row, col));
                (*(outputs[outputID].buffer))[count] = paramout;

                std::vector<float> buf(row * col * 2);
                fin.read((char *)&buf[0], buf.size() * sizeof(float));
                int actual_size = fin.gcount();
                if(!(actual_size)) {
                    isObject = false;
                } else {
                    if(actual_size != (row * col * 2 * sizeof(float))) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL/ROW values if set correctly.", __FILE__, __LINE__);
                    }
                    for (unsigned int cnt1 = 0; cnt1 < row; cnt1++) {
                        for (unsigned int cnt2 = 0; cnt2 < col; cnt2++) {
                            float tmp_r = buf[cnt2 * 2 + cnt1 * col * 2 + 0];
                            float tmp_i = buf[cnt2 * 2 + cnt1 * col * 2 + 1];
                            (*paramout)(cnt1, cnt2) = complex<float>(tmp_r, tmp_i);
                            if(ENABLE_LOG) file << (*paramout)(cnt1, cnt2) << " ";
                        }
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            default:
                break;
        }

        if(fin.eof() || isObject == false) { 
            cout << "This is the end Frame @ Count [" << count  << "]"<< endl;
            (*(outputs[noteofID].buffer))[count] = FalseObject;
        }
    } 
};

