#include <string>
#include <iostream>
#include <cstdlib>
#include <complex>
#include <cmath>

#include "BufferedNode.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"
#include "Source.h"

#include "BF.h"
#include "BF_Matrix.h"
#include "BF_ML.h"

#include "ArrayInit.h"
#include "MICARY_struct.h"
//#include <../config.h>
#include <iostream>

using namespace std;
using namespace FD;

#ifdef ENABLE_HARKIO3
	#include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
	#include "HarkFileIO.h"
	using namespace harkio2;
#elif ENABLE_HARKIO1
	#include "HarkFileIO.h"
	using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"

class ML;

DECLARE_NODE(ML);
/*Node
 *
 * @name ML
 * @category HARK:Separation
 * @description Execute sound source separation with ML.
 *
 * @input_name INPUT_FRAMES
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @input_name INPUT_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @input_name INPUT_NOISE_CM
 * @input_type Matrix<complex<float> >
 * @input_description Noise correlation matrix.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Separated spectrum with a source ID(key), and the value is the separated spectrum for each sound source (Vector<complex<float> >).
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * @parameter_name LOWER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Lower bound of frequency (Hz). [default: 0]
 * 
 * @parameter_name UPPER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 8000
 * @parameter_description Upper bound of frequency (Hz). [default: 8000]
 *
 * @parameter_name TF_CONJ_FILENAME
 * @parameter_type string
 * @parameter_description Filename of a pre-measured transfer function for separation.
 *
 * @parameter_name REG_FACTOR
 * @parameter_type float
 * @parameter_value 0.0001
 * @parameter_description Reg Factor value. [default 0.0001]
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description enable debug print
 *
 END*/
#if 0
 * @parameter_name EPS_VALUE
 * @parameter_type float
 * @parameter_value 1.0e-10
 * @parameter_description EPS_VALUE to avoid divergence . [default e-10]
#endif

class ML : public BufferedNode
{
public:
    ML(string nodeName, ParameterSet params)
	: BufferedNode(nodeName, params)
    {
		/*== FlowDesigner 入出力インタフェース ==*/

		/* 入出力IDの割り当て */
		this->input_frameID        = addInput("INPUT_FRAMES");
		this->input_signalSourceID = addInput("INPUT_SOURCES");
		this->input_noiseCMID      = addInput("INPUT_NOISE_CM");
		this->output_ID            = addOutput("OUTPUT");

		/*== パラメータの取得と検証 ==*/

		/* 周波数解析窓長パラメータ値の取得と検証 */
		this->length = dereference_cast<int>(parameters.get("LENGTH"));
		if (this->length <= 0) {
		    throw (new NodeException(NULL, "ML:LENGTH is invalid value.", __FILE__, __LINE__));
		}

		/* サンプリング周波数パラメータ値の取得と検証 */
		this->samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (this->samplingRate <= 0){
		    throw (new NodeException(NULL, "ML:SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}
		
		/* 上下限周波数の取得と検証 */
		int lower_bound_freq = dereference_cast<int>(parameters.get("LOWER_BOUND_FREQUENCY"));
		int upper_bound_freq = dereference_cast<int>(parameters.get("UPPER_BOUND_FREQUENCY"));
		if (lower_bound_freq < 0 || lower_bound_freq > this->samplingRate/2) {
	    	throw (new NodeException(NULL, "ML:LOWER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (upper_bound_freq < 0 || upper_bound_freq > this->samplingRate/2) {
	    	throw (new NodeException(NULL, "ML:UPPER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (lower_bound_freq > upper_bound_freq) {
	    	throw (new NodeException(NULL, "ML:LOWER_BOUND_FREQUENCY is larger than UPPER_BOUND_FREQUENCY.", __FILE__, __LINE__));
		}
		this->lowerfreq = floor((double)lower_bound_freq / (double)this->samplingRate * (double)this->length);
		this->upperfreq = ceil((double)upper_bound_freq / (double)this->samplingRate * (double)this->length);

#if 0	   
		/* 伝達関数の決定方法 */
		// string tf_conj = object_cast<string>(parameters.get("TF_CONJ"));
		string tf_conj = "DATABASE";
		if (tf_conj == "DATABASE") {
		    /* 伝達関数の決定方法が DATABASE である場合は、伝達関数の定義ファイルを要求する。 */
#endif
		    if (parameters.exist("TF_CONJ_FILENAME") == true) {
				this->tfFileName = object_cast<string>(parameters.get("TF_CONJ_FILENAME"));
		    } else {
				throw (new NodeException(NULL,"ML:TF_CONJ_FILENAME is not set.",__FILE__,__LINE__)) ;
		    }

#if 0 // 使用していないため削除
		    /* 音速値 */
		    this->soundSpeed = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));

		    this->tfType = TfDatabase;

		}
		else if (tf_conj == "CALC") {
		    /* 伝達関数の決定方法が CALC である場合は、マイク位置の定義ファイルを要求する。 */
		    if (parameters.exist("MIC_FILENAME") == true) {
				this->micFileName = object_cast<string>(parameters.get("MIC_FILENAME"));
		    } else {
				throw (new NodeException(NULL,"ML:MIC_FILENAME is not set.",__FILE__,__LINE__)) ;
		    }

		    /* 伝達関数の決定方法が CALC である場合は、マイク座標原点の設定を要求する。 */
		    string micPosShift = object_cast<string>(parameters.get("MIC_POS_SHIFT"));
		    if (micPosShift == "FIX") {
				this->micShiftType = MicShiftFixed;
		    } else if(micPosShift == "SHIFT") {
				this->micShiftType = MicShiftShifted;
		    } else {
				throw (new NodeException(NULL,"ML:MIC_FILENAME is not set.",__FILE__,__LINE__)) ;
		    }
		    
		    /* 音速値 */
		    this->soundSpeed = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));
			
		    this->tfType = TfCalc;

		}
		else {
		    // それ以外の値が指定された場合はエラー
		    throw (new NodeException(NULL,"Unexpected ML:TF_CONJ is set.",__FILE__,__LINE__)) ;
		}
#endif

		/* ML で用いるパラメータ */
		/* RegFactor 値 */
		if (parameters.exist("REG_FACTOR")) {
			this->regFactor = dereference_cast<float>(parameters.get("REG_FACTOR"));
		} else {
		    /* それ以外の値が指定された場合はエラー */
		    throw (new NodeException(NULL,"ML:parameter REG_FACTOR is not set.",__FILE__,__LINE__)) ;
		}

#if 0
		/* EPS_VALUE */
		if (parameters.exist("EPS_VALUE")) {
			this->epsValue = dereference_cast<float>(parameters.get("EPS_VALUE"));
		} else {
			// 存在しない場合はデフォルト値を使用する
			this->epsValue = 1.0e-10;
		}
#endif

		/* デバッグ出力用フラグ */
		if (parameters.exist("ENABLE_DEBUG")) {
			this->enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
		}
		
		/*== パラメータに基づいて初期化 ==*/

		/* 伝達関数データベース構築 */
#if 0 // 使用していないため削除
		BF::TransformFunctionDatabase newTransformFunctionDatabase;
		if (this->tfType == TfDatabase) {
		    /* HARK 形式のデータベース読み込み */
		    MICARY_TFDBFileHeader tfdbh;
		    CHarkFileIO harkFileIO;
		    if (!harkFileIO.ReadHeader(this->tfFileName.c_str(), &tfdbh)) {
				throw new NodeException(this, "Failed to open TF data file <" + this->tfFileName + ">.", __FILE__, __LINE__);
		    }

		    MICARY_TFDatabase tfdb;
		    TFDatabaseAlloc(&tfdb, tfdbh.lNFreq, tfdbh.lNSrc, tfdbh.lNMic);
		    if (!harkFileIO.ReadInfo(this->tfFileName.c_str(), &tfdbh, &tfdb.micary_locaSource, &tfdb.micary_locaMic, &tfdb.micary_iaFreqLine)
		       || !harkFileIO.ReadData(this->tfFileName.c_str(), &tfdb.micary_cmpaTF)) {
				throw new NodeException(this, "Failed to read TF data file <" + this->tfFileName + ">.", __FILE__, __LINE__);
		    }

		    /* データベース形式を変換 */
		    int numSrc  = tfdb.micary_locaSource.ha_numLocation;
		    int numMic  = tfdb.micary_locaMic.ha_numLocation;
		    int numFreq = tfdb.micary_iaFreqLine.ha_numValue;
		    for (int i = 0; i < numSrc; i++) {
				MICARY_Location &l = tfdb.micary_locaSource.micary_Location[i];
				BF::SourceLocation sl(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaSource.ha_idLocation[i], true);
				newTransformFunctionDatabase.sourceLocations.push_back(sl);
		    }
		    for (int i = 0; i < numMic; i++) {
				MICARY_Location &l = tfdb.micary_locaMic.micary_Location[i];
				BF::MicLocation ml(l.ha_fX, l.ha_fY, l.ha_fZ);
				newTransformFunctionDatabase.micLocations.push_back(ml);
		    }
		    for (int i = 0; i < numFreq; i++) {
				BF::Matrix tf(numMic, numSrc);

				for (int j = 0; j < numMic; j++) {
			    	for (int k = 0; k < numSrc; k++) {
						HA_Complex &v =tfdb.micary_cmpaTF.ha_cmpValue[(i * numSrc + k) * numMic + j];
						// 2012/05/24 added by N.Tanaka H=conj(TFCONJ)
						//tf.setElementAt(j, k, v.re, v.im);
						tf.setElementAt(j, k, v.re, -v.im);
			    	}
				}
				newTransformFunctionDatabase.transformFunctions.push_back(tf);
		    }
			TFDatabaseFree(&tfdb);
		}

		if (this->tfType == TfCalc) {
		    /* マイク位置読み込み */
		    MICARY_LocationArray micPositions;
		    CHarkFileIO harkFileIO;
		    LocationArrayAlloc(&micPositions, 0);
		    if (!harkFileIO.LoadPositionData(this->micFileName.c_str(), &micPositions, __FILE_TYPE_MIC)) {
				throw new NodeException(NULL, "Failed to Read mic position file.", __FILE__, __LINE__);
		    }

		    std::vector<BF::MicLocation> newMicPositions;
		    for (int i = 0; i < micPositions.ha_numLocation; i++) {
				const MICARY_Location &mic = micPositions.micary_Location[i];
				newMicPositions.push_back(BF::MicLocation(mic.ha_fX, mic.ha_fY, mic.ha_fZ));
		    }
		    this->micPositions.swap(newMicPositions);

		}
		else if(this->tfType == TfDatabase) {
		    this->micPositions = newTransformFunctionDatabase.micLocations;
		}
#endif
		CHarkFileIOByMatrix	cHarkMat;
		BF::TransformFunctionDatabase newTransformFunctionDatabase;
		cHarkMat.ReadTFData(this->tfFileName.c_str(), 300.0, newTransformFunctionDatabase);	// neighborは使用しないためデフォルト値を設定

#if 0 // 使用していないため削除
		/*== 分離行列初期値の初期化 ==*/
		BF::TransformFunctionDatabase newInitSeparationMatrixDatabase;
		if (iFlgW == 0) {
			int numSrc = newTransformFunctionDatabase.sourceLocations.size();
			for (int i = 0; i < numSrc; i++) {
				newInitSeparationMatrixDatabase.sourceLocations.push_back(newTransformFunctionDatabase.sourceLocations[i]);
			}
			int numMic = newTransformFunctionDatabase.micLocations.size();
			for (int i = 0; i < numMic; i++) {
				newInitSeparationMatrixDatabase.micLocations.push_back(newTransformFunctionDatabase.micLocations[i]);
			}
			int numFreq = newTransformFunctionDatabase.transformFunctions.size();
			for (int i = 0; i < numFreq; i++) {
				newInitSeparationMatrixDatabase.transformFunctions.push_back(newTransformFunctionDatabase.transformFunctions[i]);
			}
		}
		else {
		}
#endif

		/* adaptBF 作業変数確保 */
#if 0
		this->bfAdapter = new BF::CML(this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->regFactor, this->epsValue);
#endif
		this->bfAdapter = new BF::CML(this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->regFactor);

		/* 伝達関数データベースを登録 */
#if 0
		if (this->tfType == TfDatabase) {
#endif
		    this->bfAdapter->setTransformFunctionDatabase(newTransformFunctionDatabase);
#if 0
		}
#endif
    }

    virtual ~ML()
    {
		delete this->bfAdapter;
    }

public:
    virtual int
    translateInput(const string &inputName)
    {
		if (inputName == "INPUT_FRAMES") {
			return this->input_frameID;
		} else if(inputName == "INPUT_SOURCES") {
			return this->input_signalSourceID;
		} else if(inputName == "INPUT_NOISE_CM") {
			return this->input_noiseCMID;
		} else {
			throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
		}
    }

    void calculate(int output_id, int count, Buffer &out)
    {
		/*== 入出力の準備 ==*/

		/* スペクトル入力 */
		ObjectRef orInputTemp = getInput(this->input_frameID, count);
		Matrix<complex<float> > &inputValueMatrix = object_cast<Matrix<complex<float> > >(orInputTemp);
	
		int numMic  = inputValueMatrix.nrows();	// チャンネル数
		int numFreq = inputValueMatrix.ncols();	// 周波数ライン数
	    	    
		/* 信号音源位置情報入力 */
		RCPtr<Vector<ObjectRef> > pSignalSources = getInput(this->input_signalSourceID, count);
		const Vector<ObjectRef> *signalSources = pSignalSources.get();
		int numSource = signalSources->size(); // 信号音源数
	    
		/* ノイズ相関行列入力 */
		if (this->input_noiseCMID >= 0) {
			// CMMakerFromFFTの出力の型 Matrix<complex<float> >
			// CMMakeFromFFTの出力 .nrows() は numFreq, 
			//                     .ncols() は numMic * numMic

			RCPtr<Matrix<complex<float> > > pNoiseCM = getInput(this->input_noiseCMID, count);

			if (numFreq != pNoiseCM->nrows() || (numMic * numMic) != pNoiseCM->ncols()) {
				throw new NodeException(this, "ML: size of noise correlation matrix is defferent from input.", __FILE__, __LINE__);
			}

//std::cout << "[ML] count:" << count << std::endl;
			if (count == 0) {
				/* 初期化 */
				this->bfAdapter->covarianceMatrices.resize(0);
				for (int k = 0; k < numFreq; k++) {
					BF::Matrix rx(numMic, numMic);
					for (int i = 0; i < numMic; i++) {
						for (int j = 0; j < numMic; j++) {
							rx.setElementAt(i, j, 0.0f, 0.0f);
						}
					}
					this->bfAdapter->covarianceMatrices.push_back(rx);
				}
			}

			for (int k = 0; k < numFreq; k++) {
				BF::Matrix rx(numMic, numMic);
				for (int i = 0; i < numMic; i++) {
					for (int j = 0; j < numMic; j++) {
						rx.setElementAt(i, j, (*pNoiseCM)(k, i * numMic + j));
					}
				}
				this->bfAdapter->covarianceMatrices[k] = rx;
			}
		}
		else {
			std::cout << "When you use ML, you must input noise correlation matrix to INPUT_NOISE_CM." <<  std::endl;
			throw new NodeException(this, "ML needs INPUT_NOISE_CM", __FILE__, __LINE__);
		}

		/* 出力スペクトルの領域を準備 */
		ObjectRef orOutputTemp = ObjectRef(new Map<int,ObjectRef>);
		Map<int, ObjectRef> &mpOutputValue = object_cast<Map<int,ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;
		mpOutputValue.clear();

		/*== 処理用データの準備 ==*/

		/* 入力スペクトル */
		BF::Matrix inputSpectrum(numMic, numFreq);
		for (int i = 0 ; i < numMic; i++) {
			for (int j = 0; j < numFreq; j++) {
				complex<float> v = inputValueMatrix(i, j);
				inputSpectrum.setElementAt(i, j, BF::Complex(v.real(), v.imag()));
			}
		}

		/* 出力スペクトル */
		BF::Matrix outputSpectrum(numSource, numFreq);

		/* 音源位置 */
		std::vector<BF::SourceLocation> sourcePositions;
		for (int i = 0 ; i < numSource; i++) {
		    RCPtr<Source> src = signalSources->at(i);
			if (src->compare_mode == Source::TFINDEX){
			    sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], src->tfindex[0], src->id, true));
			}
			else { // Source::DEG
			    sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], -1, src->id, true));
			    int idx = this->bfAdapter->transformFunctionDatabase->findNearestSourceIndex(sourcePositions[sourcePositions.size()-1]);
			    sourcePositions[sourcePositions.size()-1].id = this->bfAdapter->transformFunctionDatabase->sourceLocations[idx].id;

			}
//std::cout << sourcePositions[sourcePositions.size()-1].x << "," << sourcePositions[sourcePositions.size()-1].y << "," << sourcePositions[sourcePositions.size()-1].z << "," << sourcePositions[sourcePositions.size()-1].id << "," << sourcePositions[sourcePositions.size()-1].sourceID << std::endl;
		}

		if (numSource > 0) {
			/*== 伝達関数適用（更新または継承）および分離行列初期化 ==*/
			this->bfAdapter->updateSources(sourcePositions);

			/*== ML 実行 ==*/
			this->bfAdapter->executeBeamForming(inputSpectrum, outputSpectrum);

			/*== 分離行列適応処理 == →なし*/
			//this->bfAdapter->adaptSeparationMatrices();

			/*== 処理結果を出力スペクトルの領域へ格納 ==*/
			for (int i = 0 ; i < numSource; i++) {
				int id = sourcePositions[i].sourceID;
				ObjectRef orOutputTempV = ObjectRef(Vector<complex<float> >::alloc(numFreq));
				Vector<complex<float> > &vcfOutputValue = object_cast<Vector<complex<float> > > (orOutputTempV) ;
				
				for (int j = 0 ; j < numFreq; j++) {
					const BF::Complex v = outputSpectrum.getElementAt(i, j);
					float re = v.real();
					float im = v.imag();
					if ((fpclassify(re) == FP_NAN || fpclassify(im) == FP_NAN) && (this->enable_debug)) {
						std::cout << "ML: Output Spectrum data ["<< i <<" , " <<  j << "] is NAN. Please check the parameters." << endl ;
					}
					if ((fpclassify(re) == FP_INFINITE || fpclassify(im) == FP_INFINITE) && (this->enable_debug)) {
						std::cout << "ML: Output Spectrum data ["<< i <<" , " <<  j << "] is INF. Please check the parameters." << std::endl ;
					}
					vcfOutputValue[j] = complex<float>(re, im);
				}
				mpOutputValue.insert(make_pair(id,orOutputTempV));
			}
		}
    }

#if 0 // HarkFileIOMatrixを使用するため
    void
    exportSeparationMatrices(const string              &filename,          /* 出力先ファイル名 */
			     int                                   numFreq,            /* 周波数ビンの数 */
			     int                                   samplingRate,       /* サンプリング周波数 */
			     const std::vector<BF::MicLocation>    &micPositions,      /* マイク位置 */
			     const std::vector<BF::SourceLocation> &sourcePositions,   /* 音源位置 */
			     const std::vector<BF::Matrix>         &separationMatrices /* 最新の分離行列群 */)
    {
	int numMic = micPositions.size();
	int numSrc = sourcePositions.size();

	MICARY_SMDatabase smDatabase = {0};
		
	SMDatabaseAlloc(&smDatabase, 0, 0, 0);
	SMDatabaseAlloc(&smDatabase, numFreq, numSrc, numMic);

	for(int i = 0; i < numMic; i++){
	    const BF::MicLocation &ml = micPositions[i];
	    smDatabase.micary_locaMic.ha_idLocation[i] = i;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fX = ml.x;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fY = ml.y;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fZ = ml.z;
	    smDatabase.micary_locaMic.micary_Location[i].ha_idPos = i;
	}

	for(int i = 0; i < numSrc; i++){
	    const BF::SourceLocation &sl = sourcePositions[i];
	    smDatabase.micary_locaSource.ha_idLocation[i] = sl.sourceID;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fX = sl.x;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fY = sl.y;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fZ = sl.z;
	    smDatabase.micary_locaSource.micary_Location[i].ha_idPos = sl.sourceID;
	}

	for(int i = 0 ; i < numFreq; i++){
	    smDatabase.micary_smSeparation.micary_iaFreqLinex.ha_iValue[i] = (double)samplingRate * (i * 2 + 1) / 4 / numFreq;
	}

	int c = 0;
	for(int i = 0 ; i < numFreq; i++){
	    const BF::Matrix &mat = separationMatrices[i];

	    for(int j = 0 ; j < numSrc; j++){
		for(int k = 0 ; k < numMic; k++){
		    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].re = mat.getElementAt(j, k).real();
		    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].im = mat.getElementAt(j, k).imag();
		    c++;
		}
	    }
	}

	MICARY_TFDBFileHeader tfdbfh;
	tfdbfh.lFs       = samplingRate;
	tfdbfh.lNFFT     = numFreq * 2;
	tfdbfh.lNMic     = numMic;
	tfdbfh.lNSrc     = numSrc;
	tfdbfh.lNFreq    = numFreq;
	tfdbfh.iVersion  = 1;
	tfdbfh.iFileType = FORMAT_TYPE_W;

	CHarkFileIO cHark;
	if(!cHark.WriteData(filename.c_str(), &tfdbfh,
			    smDatabase.micary_smSeparation.micary_cmpaSeparation,
			    smDatabase.micary_smSeparation.micary_iaFreqLinex,
			    smDatabase.micary_locaSource,
			    smDatabase.micary_locaMic,
			    "",
			    "",
			    "",
			    "",
			    "",
			    "GHDSS_EXPORT")){
	    SMDatabaseFree(&smDatabase);
	    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
	}

	SMDatabaseFree(&smDatabase);
    }
#endif
    void
    exportSeparationMatrices(const string  &filename, /* 出力先ファイル名 */
							 int samplingRate,        /* サンプリング周波数 */
							 int numFreq,             /* 周波数ビンの数 */
    						 BF::Database *SeparationMatricesDatabase)
    {
		CHarkFileIOByMatrix	cHarkMat;
		if (!cHarkMat.WriteWData(filename.c_str(), samplingRate, numFreq, *(SeparationMatricesDatabase))) {
		    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
		}
	}


private:
    /*=== FlowDesigner 入出力 ===*/

    /* フレームデータの入力ID */
    int input_frameID;

    /* 信号音源情報の入力ID */
    int input_signalSourceID;

    /* ノイズ相関行列の入力ID */
    int input_noiseCMID;

    /* 処理結果フレームデータの出力ID */
    int output_ID;

    /*=== パラメータ ===*/

#if 0
    /* 伝達関数をファイルから読み込んで使用するか？ */
    enum TfType {
	TfUnknown = 0, TfDatabase, TfCalc,
    };
    TfType tfType;
#endif

    /* 伝達関数をファイルから読み込んで使用する場合のファイル名 */
    string tfFileName;

#if 0
    /* マイク原点位置の設定 */
    enum MicShiftType {
	MicShiftUnknown = 0, MicShiftFixed, MicShiftShifted,
    };
    MicShiftType micShiftType;

    /* マイク位置の定義ファイル名 */
    string micFileName;
#endif

    /* 周波数解析窓長 */
    int length;

    /* サンプリング周波数 */
    int samplingRate;

    /* 音速値 */
    float soundSpeed;

    /* 分析対象周波数 */
    int lowerfreq;
    int upperfreq;
    
    /* ML で用いる RegFactor 値 */
    float regFactor;

#if 0
    /* ML で用いる epsValue 値 */
    float epsValue;
#endif

    /* デバッグ出力用フラグ */
    bool enable_debug;

    /*=== 作業用変数 ===*/

#if 0
    /* マイク位置 (tfType == TfCalc 時のみ) */
    std::vector<BF::MicLocation> micPositions;
#endif

    /* 分離行列の更新を行う処理オブジェクト  */
    BF::BF *bfAdapter;

};

