#include <string>
#include <iostream>
#include <cstdlib>
#include <complex>
#include <cmath>

#include "BufferedNode.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"
#include "Source.h"

#include "BF.h"
#include "BF_Matrix.h"
#include "BF_MVDR.h"

#include "ArrayInit.h"
#include "MICARY_struct.h"
//#include <../config.h>
#include <iostream>

using namespace std;
using namespace FD;

#ifdef ENABLE_HARKIO3
	#include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
	#include "HarkFileIO.h"
	using namespace harkio2;
#elif ENABLE_HARKIO1
	#include "HarkFileIO.h"
	using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"

class MVDR;

DECLARE_NODE(MVDR);
/*Node
 *
 * @name MVDR
 * @category HARK:Separation
 * @description Execute sound source separation with MVDR.
 *
 * @input_name INPUT_FRAMES
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @input_name INPUT_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @input_name INPUT_NOISE_CM
 * @input_type Matrix<complex<float> >
 * @input_description Noise correlation matrix.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Separated spectrum with a source ID(key), and the value is the separated spectrum for each sound source (Vector<complex<float> >).
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * @parameter_name LOWER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Lower bound of frequency (Hz). [default: 0]
 * 
 * @parameter_name UPPER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 8000
 * @parameter_description Upper bound of frequency (Hz). [default: 8000]
 *
 * @parameter_name TF_CONJ_FILENAME
 * @parameter_type string
 * @parameter_description Filename of a pre-measured transfer function for separation.
 *
 * @parameter_name REG_FACTOR
 * @parameter_type float
 * @parameter_value 0.001
 * @parameter_description Reg Factor value. [default 0.001]
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description enable debug print
 *
 END*/

class MVDR : public BufferedNode
{
public:
    MVDR(string nodeName, ParameterSet params)
	: BufferedNode(nodeName, params)
    {
		/*== FlowDesigner 入出力インタフェース ==*/

		/* 入出力IDの割り当て */
		this->input_frameID        = addInput("INPUT_FRAMES");
		this->input_signalSourceID = addInput("INPUT_SOURCES");
		this->input_noiseCMID      = addInput("INPUT_NOISE_CM");
		this->output_ID            = addOutput("OUTPUT");

		/*== パラメータの取得と検証 ==*/

		/* 周波数解析窓長パラメータ値の取得と検証 */
		this->length = dereference_cast<int>(parameters.get("LENGTH"));
		if (this->length <= 0) {
		    throw (new NodeException(NULL, "MVDR:LENGTH is invalid value.", __FILE__, __LINE__));
		}

		/* サンプリング周波数パラメータ値の取得と検証 */
		this->samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (this->samplingRate <= 0) {
		    throw (new NodeException(NULL, "MVDR:SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}
		
		/* 上下限周波数の取得と検証 */
		int lower_bound_freq = dereference_cast<int>(parameters.get("LOWER_BOUND_FREQUENCY"));
		int upper_bound_freq = dereference_cast<int>(parameters.get("UPPER_BOUND_FREQUENCY"));
		if (lower_bound_freq < 0 || lower_bound_freq > this->samplingRate/2){
	    	throw (new NodeException(NULL, "MVDR:LOWER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (upper_bound_freq < 0 || upper_bound_freq > this->samplingRate/2){
	    	throw (new NodeException(NULL, "MVDR:UPPER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (lower_bound_freq > upper_bound_freq){
	    	throw (new NodeException(NULL, "MVDR:LOWER_BOUND_FREQUENCY is larger than UPPER_BOUND_FREQUENCY.", __FILE__, __LINE__));
		}
		this->lowerfreq = floor((double)lower_bound_freq / (double)this->samplingRate * (double)this->length);
		this->upperfreq = ceil((double)upper_bound_freq / (double)this->samplingRate * (double)this->length);
	   
		/* 伝達関数の定義ファイル */
	    if (parameters.exist("TF_CONJ_FILENAME") == true) {
			this->tfFileName = object_cast<string>(parameters.get("TF_CONJ_FILENAME"));
	    } else {
			throw (new NodeException(NULL,"MVDR:TF_CONJ_FILENAME is not set.",__FILE__,__LINE__)) ;
	    }

		/* RegFactor 値 */
		this->regFactor = dereference_cast<float>(parameters.get("REG_FACTOR"));
		if (this->regFactor <= 0) {
		    throw (new NodeException(NULL, "MVDR:REG_FACTOR is invalid value.", __FILE__, __LINE__));
		}

		/* デバッグ出力用フラグ */
		if (parameters.exist("ENABLE_DEBUG")) {
		  this->enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
		}
		
		/*== パラメータに基づいて初期化 ==*/

		/* 伝達関数データベース構築 */
		CHarkFileIOByMatrix	cHarkMat;
		BF::TransformFunctionDatabase newTransformFunctionDatabase;
		cHarkMat.ReadTFData(this->tfFileName.c_str(), 300.0, newTransformFunctionDatabase);	// neighborは使用しないためデフォルト値を設定

		/* adaptBF 作業変数確保 */
		this->bfAdapter = new BF::CMVDR(this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->regFactor); 

		/* 伝達関数データベースを登録 */
	    this->bfAdapter->setTransformFunctionDatabase(newTransformFunctionDatabase);
    }

    virtual ~MVDR()
    {
		delete this->bfAdapter;
    }

public:
    virtual int
    translateInput(const string &inputName)
    {
		if (inputName == "INPUT_FRAMES") {
			return this->input_frameID;
		} else if (inputName == "INPUT_SOURCES") {
			return this->input_signalSourceID;
		} else if (inputName == "INPUT_NOISE_CM") {
			return this->input_noiseCMID;
		} else {
			throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
		}
    }

    void calculate(int output_id, int count, Buffer &out)
    {
		/*== 入出力の準備 ==*/

		/* スペクトル入力 */
		ObjectRef orInputTemp = getInput(this->input_frameID, count);
		Matrix<complex<float> > &inputValueMatrix = object_cast<Matrix<complex<float> > >(orInputTemp);

		int numMic  = inputValueMatrix.nrows();	// チャンネル数
		int numFreq = inputValueMatrix.ncols();	// 周波数ライン数
					
		/* 信号音源位置情報入力 */
		RCPtr<Vector<ObjectRef> > pSignalSources = getInput(this->input_signalSourceID, count);
		const Vector<ObjectRef> *signalSources = pSignalSources.get();
		int numSource = signalSources->size(); // 信号音源数
		
		/* ノイズ相関行列入力 */
		if (this->input_noiseCMID >= 0) {
			// CMMakerFromFFTの出力の型 Matrix<complex<float> >
			// CMMakeFromFFTの出力 .nrows() は numFreq, 
			//                     .ncols() は numMic * numMic

			RCPtr<Matrix<complex<float> > > pNoiseCM = getInput(this->input_noiseCMID, count);

			if (numFreq != pNoiseCM->nrows() || (numMic * numMic) != pNoiseCM->ncols()) {
				throw new NodeException(this, "MVDR: size of noise correlation matrix is defferent from input.", __FILE__, __LINE__);
			}

//std::cout << "[MVDR] count:" << count << std::endl;
			if (count == 0) {
				/* 初期化 */
				this->bfAdapter->covarianceMatrices.resize(0);
				for (int k = 0; k < numFreq; k++) {
					BF::Matrix rx(numMic, numMic);
					for (int i = 0; i < numMic; i++) {
						for (int j = 0; j < numMic; j++) {
							rx.setElementAt(i, j, 0.0f, 0.0f);
						}
					}
					this->bfAdapter->covarianceMatrices.push_back(rx);
				}
			}

			for (int k = 0; k < numFreq; k++) {
				BF::Matrix rx(numMic, numMic);
				for (int i = 0; i < numMic; i++) {
					for (int j = 0; j < numMic; j++) {
						rx.setElementAt(i, j, (*pNoiseCM)(k, i * numMic + j));
					}
				}
				this->bfAdapter->covarianceMatrices[k] = rx;
			}
		}
		else {
			std::cout << "When you use MVDR, you must input noise correlation matrix to INPUT_NOISE_CM." <<  std::endl;
			throw (new NodeException(this, "MVDR needs INPUT_NOISE_CM", __FILE__, __LINE__));
		}

		/* 出力スペクトルの領域を準備 */
		ObjectRef orOutputTemp = ObjectRef(new Map<int,ObjectRef>);
		Map<int, ObjectRef> &mpOutputValue = object_cast<Map<int,ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;
		mpOutputValue.clear();

		/*== 処理用データの準備 ==*/

		/* 入力スペクトル */
		BF::Matrix inputSpectrum(numMic, numFreq);
		for (int i = 0 ; i < numMic; i++) {
			for (int j = 0; j < numFreq; j++) {
				complex<float> v = inputValueMatrix(i, j);
				inputSpectrum.setElementAt(i, j, BF::Complex(v.real(), v.imag()));
			}
		}

		/* 入力パワースペクトル */
		BF::Matrix pw(numMic, numFreq);
		for (int i = 0; i < numMic; i++) {
			for (int j = 0; j < numFreq; j++) {
				double v = std::abs(inputValueMatrix(i, j));
				pw.setElementAt(i, j, v * v);
			}
		}
		this->bfAdapter->inputPowerSpectrum = pw;

		/* 出力スペクトル */
		BF::Matrix outputSpectrum(numSource, numFreq);

		/* 音源位置 */
		std::vector<BF::SourceLocation> sourcePositions;
		for (int i = 0 ; i < numSource; i++) {
		    RCPtr<Source> src = signalSources->at(i);
			if (src->compare_mode == Source::TFINDEX) {
	        	sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], src->tfindex[0], src->id, true));
			}
			else { // Source::DEG
	        	sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], -1, src->id, true));
				int idx = this->bfAdapter->transformFunctionDatabase->findNearestSourceIndex(sourcePositions[sourcePositions.size()-1]);
				sourcePositions[sourcePositions.size()-1].id = this->bfAdapter->transformFunctionDatabase->sourceLocations[idx].id;
			}
		}

		if (numSource > 0) {
			/*== 伝達関数適用（更新または継承）および分離行列初期化 ==*/
			this->bfAdapter->updateSources(sourcePositions);

			/*== MVDR 実行 ==*/
			this->bfAdapter->executeBeamForming(inputSpectrum, outputSpectrum);

			/*== 分離行列適応処理 == →なし*/
			//this->bfAdapter->adaptSeparationMatrices();

			/*== 処理結果を出力スペクトルの領域へ格納 ==*/
			for (int i = 0 ; i < numSource; i++) {
				int id = sourcePositions[i].sourceID;
				ObjectRef orOutputTempV = ObjectRef(Vector<complex<float> >::alloc(numFreq));
				Vector<complex<float> > &vcfOutputValue = object_cast<Vector<complex<float> > > (orOutputTempV) ;
			
				for (int j = 0 ; j < numFreq; j++) {
					const BF::Complex v = outputSpectrum.getElementAt(i, j);
					float re = v.real();
					float im = v.imag();
					if ((fpclassify(re) == FP_NAN || fpclassify(im) == FP_NAN) && (this->enable_debug)) {
						std::cout << "MVDR: Output Spectrum data ["<< i <<" , " <<  j << "] is NAN. Please check the parameters." << std::endl ;
					}
					if ((fpclassify(re) == FP_INFINITE || fpclassify(im) == FP_INFINITE) && (this->enable_debug)) {
						std::cout << "MVDR: Output Spectrum data ["<< i <<" , " <<  j << "] is INF. Please check the parameters." << std::endl ;
					}
					vcfOutputValue[j] = complex<float>(re, im);
				}
				mpOutputValue.insert(make_pair(id,orOutputTempV));
			}
		}
	}

    void exportSeparationMatrices(const string  &filename,	/* 出力先ファイル名 */
							      int samplingRate,			/* サンプリング周波数 */
							      int numFreq,				/* 周波数ビンの数 */
    						      BF::Database *SeparationMatricesDatabase)
    {
		CHarkFileIOByMatrix	cHarkMat;
		if (!cHarkMat.WriteWData(filename.c_str(), samplingRate, numFreq, *(SeparationMatricesDatabase))) {
		    throw (new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__));
		}
	}


private:
    /*=== FlowDesigner 入出力 ===*/

    /* フレームデータの入力ID */
    int input_frameID;

    /* 信号音源情報の入力ID */
    int input_signalSourceID;

    /* ノイズ相関行列の入力ID */
    int input_noiseCMID;

    /* 処理結果フレームデータの出力ID */
    int output_ID;

    /*=== パラメータ ===*/

    /* 伝達関数をファイルから読み込んで使用する場合のファイル名 */
    string tfFileName;

    /* 周波数解析窓長 */
    int length;

    /* サンプリング周波数 */
    int samplingRate;

    /* 音速値 */
    float soundSpeed;

    /* 分析対象周波数 */
    int lowerfreq;
    int upperfreq;
    
    /* MVDR で用いる RegFactor 値 */
    float regFactor;

    /* デバッグ出力用フラグ */
    bool enable_debug;

    /*=== 作業用変数 ===*/

    /* 分離行列の更新を行う処理オブジェクト  */
    BF::BF *bfAdapter;

};


