/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"

using namespace std;
using namespace FD;

class MapIDOffset;

DECLARE_NODE(MapIDOffset)
/*Node
 *
 * @name MapIDOffset
 * @category HARK:MISC
 * @description ID_OFFSET parameter can be offset source ID values of INPUT Map.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Vector<float> or Vector<complex<float> > or Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The key is offset source ID, and the value is the same as INPUT.
 *
 * @parameter_name ID_OFFSET
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The key starting value of OUTPUT.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

class MapIDOffset : public BufferedNode {
	int inputID;
	int outputID;
	int const_id;
	bool debug;

public:
	MapIDOffset(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		const_id = dereference_cast<int>(parameters.get("ID_OFFSET"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		if (debug == true) cout << "******* MapIDOffset debug point-Input check (count=" << count << ")" << endl;

		ObjectRef input = getInput(inputID, count);
		if (typeid(*input) == typeid(Map<int, ObjectRef>)) {
			RCPtr<Map<int, ObjectRef> > inp = input;
			const Map<int, ObjectRef>& in = *inp;

			if (debug == true) {
				if (in.empty() == true) {
					cout << "map empty" << endl;
				}
				else {
					cout << "Input map begins from " << (in.begin()->first) << endl;
					cout << "Number of channels: " << in.size() << endl;
					cout << "Type of first channel of ObjectRef of input map is: ";
					if (typeid(*(in.begin()->second)) == typeid(Vector<float>)) {
						cout << "Vector<float>";
						cout << ", size=" << (object_cast<Vector<float> >(in.begin()->second)).size() << endl;
					}
					else if (typeid(*(in.begin()->second)) == typeid(Vector<complex<float> >)) {
						cout << "Vector<complex<float> >";
						cout << ", size=" << (object_cast<Vector<complex<float> > >(in.begin()->second)).size() << endl;
					}
					else if (typeid(*(in.begin()->second)) == typeid(Matrix<float>)) {
						cout << "Matrix<float>";
						cout << ", rows=" << (object_cast<Matrix<float> >(in.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<float> >(in.begin()->second)).ncols() << endl;
					}
					else if (typeid(*(in.begin()->second)) == typeid(Matrix<complex<float> >)) {
						cout << "Matrix<complex<float> >";
						cout << ", rows=" << (object_cast<Matrix<complex<float> > >(in.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<complex<float> > >(in.begin()->second)).ncols() << endl;
					}
					else {
						cout << "unidentified." << endl;
					}
				}
			}

			Map<int, ObjectRef>& output = *new Map<int, ObjectRef>;
			out[count] = &output;

			if (in.empty() == false) {
				for (Map<int, ObjectRef>::const_iterator it = in.begin(); it != in.end(); ++it) {
					output[(it->first) + const_id] = it->second;

					if (debug == true) cout << "Input Source ID = " << it->first << ", Output Source ID = " << (it->first) + const_id << endl;
				}
			}
		}
		else {
			cerr << "******* MapIDOffset: Only Map<int, ObjectRef> is supported." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MapIDOffset]", __FILE__, __LINE__));
		}

		if (debug == true) cout << "******* MapIDOffset debug point-Finish" << endl;
	}

	NO_ORDER_NODE_SPEEDUP(MapIDOffset);
};
