/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <climits>

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include <math.h>
#include <csignal>
#include <string>
#include <Matrix.h>
#include "Time.h"
#include "TimeStamp.h"
#include <sys/time.h>

//#include "../config.h"

#include <ALSARecorder.hpp>
#if ENABLE_TDBD8CSUSB
#include <ASIORecorder.hpp>
#endif
#if ENABLE_TDBD16ADUSB
#include <SINICHRecorder.hpp>
#endif
#if ENABLE_RASP2
#include <RASPRecorder.hpp>
#endif
#if ENABLE_WS
#include <WSRecorder.hpp>
#endif
#if ENABLE_ASRASP
#include <ASRASPRecorder.hpp>
#endif
#if ENABLE_RASP24
#include <RASP24Recorder.hpp>
#endif
#if ENABLE_HARKNETAPI
#include <HARKNETAPIRecorder.hpp>
#endif
#if ENABLE_RASPLC
#include <RASPLCRecorder.hpp>
#endif
#if ENABLE_KANEKO
#include <KANEKORecorder.hpp>
#endif

using namespace std;
using namespace FD;
using namespace std::chrono;

class MultiAudioStreamFromMic;

DECLARE_NODE(MultiAudioStreamFromMic);
/*Node
 *
 * @name MultiAudioStreamFromMic
 * @category HARK:AudioIO
 * @description Capture sound data synchronously from multiple microphones.
 *
 * @output_name NOT_EOF
 * @output_type bool
 * @output_description Always true, this is for "condition" setting
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name CHANNEL_COUNT
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description The number of channels.
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * @parameter_name DEVICETYPE
 * @parameter_type string
 * @parameter_list ALSA:RASP:WS:TDBD16ADUSB:RASP24-16:RASP24-32:RASP-LC:KANEKO:NETWORK
 * @parameter_value WS
 * @parameter_description Device type [default: WS].
 *
 * @parameter_name GAIN
 * @parameter_type string
 * @parameter_list 0dB:12dB:24dB:36dB:48dB
 * @parameter_value 0dB
 * @parameter_valid DEVICETYPE=RASP24-16:RASP24-32
 * @parameter_description  capture gain (dB)  [default: 0dB].
 *
 * @parameter_name DEVICE
 * @parameter_type string
 * @parameter_value /dev/null
 * @parameter_description List of device name or IP address, separated with whitespace. [default: /dev/null]
 *
 * @parameter_name FRAME_COUNT_SKEW_TOLERANCE
 * @parameter_type float
 * @parameter_value 5.0
 * @parameter_description Accept frame count skewness upto this value.
 *
 END*/

bool recorder_stop_flag3 = false;

void signal_handler3(int s)
{
    recorder_stop_flag3 = true;
}

class MultiAudioStreamFromMic: public BufferedNode
{

//    vector<int> audioIDs;
    int eofID;
    int timeID;
    int timestampID;
    int length;
    int advance;
    int channel_count;
    int sampling_rate;
    string device_type;
//    Vector<string> devices;
    string device_gain;
    string dv_gain;
    int frame_count_skew_tolerance;

//    vector<Recorder *> recorders;
    bool hasTimestamp;
    int overlapped_usec;
    int sleep_read_usec;
    float amp_max;
    float amp_min;

//    vector<vector<float> > buffers;

    struct Device {
        int       outputID;
        string    deviceName;
        Recorder *recorder;
        vector<float> buffer;
	size_t        frameOffset; /* 処理開始（デバイスオープン）タイミングのずれによる cur_time 値の差を補償するため、count=0 時点での cur_time を保存しておく */
    };
    vector<Device> devices;

public:

    MultiAudioStreamFromMic(string nodeName, ParameterSet params) :
        BufferedNode(nodeName, params), devices(), hasTimestamp(false), overlapped_usec(0), amp_max(SHRT_MAX), amp_min(SHRT_MIN)
    {
        eofID = addOutput("NOT_EOF");
        timeID = addOutput("TIME");
        timestampID = addOutput("TIMESTAMP");
        if (parameters.exist("LENGTH")) {
            length = dereference_cast<int>(parameters.get("LENGTH"));
        }
        if (parameters.exist("ADVANCE")) {
            advance = dereference_cast<int>(parameters.get("ADVANCE"));
        } else {
            advance = length;
        }

        channel_count = dereference_cast<int>(parameters.get("CHANNEL_COUNT"));
        sampling_rate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
        device_type = object_cast<String>(parameters.get("DEVICETYPE"));
        device_gain = object_cast<String>(parameters.get("GAIN"));
        frame_count_skew_tolerance = (int)(dereference_cast<float>(parameters.get("FRAME_COUNT_SKEW_TOLERANCE")) * sampling_rate);

        String devicenames = object_cast<String>(parameters.get("DEVICE"));
//	cout << devices;

        overlapped_usec = (length - advance) * 1000000 / sampling_rate;
        sleep_read_usec = advance / 2 * 1000000 / sampling_rate;

        signal(SIGINT, signal_handler3);
        signal(SIGHUP, signal_handler3);

        size_t device_index;
        std::stringstream ss(devicenames);
        string deviceName;
        for(device_index = 0; std::getline(ss, deviceName, ' '); device_index++) {
            Recorder *recorder = NULL;
            if (device_type == "ALSA") {
                // device = "plughw:0";
                recorder = ALSARecorder::GetInstance();
                recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate, length * 1024);
            }
#if ENABLE_TDBD8CSUSB
            else if (device_type == "ASIO") {
                // device = "plughw:0";
                recorder = ASIORecorder::GetInstance();
                recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate,
                                     length * 1024);
            }
#endif
#if ENABLE_TDBD16ADUSB
            else if (device_type == "TDBD16ADUSB") {
                // device = "plughw:0";
                recorder = SINICHRecorder::GetInstance();
                recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate,
                                     length * 2048);
            }
#endif
#if ENABLE_RASP2
            else if (device_type == "RASP") {
                // device = "/dev/rasp/adc0";
                recorder = RASPRecorder::GetInstance();
                recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate,
                                     length * 1024);
            }
#endif
#if ENABLE_WS
            else if (device_type == "WS") {
                // device = "192.168.32.228";
                recorder = WSRecorder::GetInstance();
                int ret = recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate, length * 1024);
                if (ret == -1) {
                    throw new NodeException(NULL, string("Device can not be initialized."),
                                            __FILE__, __LINE__);
                }
                hasTimestamp = true;
                if (channel_count == 8 || channel_count == 10) {
                    amp_max = 8388607;
                    amp_min = -8388608;
                }
            }
#endif
#if ENABLE_ASRASP
            else if (device_type == "ASRASP") {
                recorder = ASRASPRecorder::GetInstance();
                int ret = recorder->Initialize(deviceName.c_str(), channel_count,
                                               sampling_rate, length * 1024);
                if (ret == -1) {
                    throw new NodeException(NULL, string("Device can not be initialized."),
                                            __FILE__, __LINE__);
                }
                hasTimestamp = true;
                if (channel_count >= 8 && channel_count <= 10) {
                    amp_max = 8388607;
                    amp_min = -8388608;
                }
            }
#endif
#if ENABLE_RASP24
            else if (device_type == "RASP24-16" || device_type == "RASP24-32") {

                string tmp_device;

                recorder = RASP24Recorder::GetInstance();

                dv_gain = "0";

                if (strcmp("0dB", device_gain.c_str()) == 0) {
                    dv_gain = "0";
                }
                if (strcmp("12dB", device_gain.c_str()) == 0) {
                    dv_gain = "1";
                } else if (strcmp("24dB", device_gain.c_str()) == 0) {
                    dv_gain = "2";
                } else if (strcmp("36dB", device_gain.c_str()) == 0) {
                    dv_gain = "3";
                } else if (strcmp("48dB", device_gain.c_str()) == 0) {
                    dv_gain = "4";
                }

                recorder = RASP24Recorder::GetInstance();

                tmp_device = deviceName + "," + device_type + "," + dv_gain;

                int ret = recorder->Initialize(tmp_device.c_str(), channel_count, sampling_rate, length * 1024);

                if (ret == -1) {

                    throw new NodeException(NULL, string("Device can not be initialized."), __FILE__, __LINE__);
                }

                hasTimestamp = true;
                amp_max = 8388607;
                amp_min = -8388608;

            }
#endif
#if ENABLE_HARKNETAPI
            else if (device_type == "NETWORK") {

                recorder = HARKNETAPIRecorder::GetInstance();

                int ret = recorder->Initialize(deviceName.c_str(), channel_count, sampling_rate, length * 1024);

                if (ret == -1) {
                    throw new NodeException(NULL, string("Device can not be initialized."), __FILE__, __LINE__);
                }

                hasTimestamp = true;

                amp_max = 2147483647;      //32bit整数時振の幅データMAX
                amp_min = -2.147483648e+09; //32bit整数時の振幅データMin
                //		amp_max =  8388607;         //24bit整数時の振幅データMAX
                //		amp_min = -8388608;         //24bit整数時の振幅データMin
                //		cout<< "HARKNETAPI END" << endl;

            }
#endif
#if ENABLE_RASPLC
            else if (device_type == "RASP-LC") {

                recorder = RASPLCRecorder::GetInstance();

                channel_count = LC_CH_NUM;
                sampling_rate = CLOCK_FREQ_16K;

                int ret = recorder->Initialize( deviceName.c_str(),
                                                channel_count,
                                                sampling_rate,
                                                length * 1024);

                if (ret == -1) {
                    throw new NodeException(NULL, string("Device can not be initialized."),__FILE__, __LINE__);
                }

                hasTimestamp = true;

                amp_max = 2147483647;      //32bit整数時振の幅データMAX
                amp_min = -2.147483648e+09;//32bit整数時の振幅データMin
                //		amp_max =  8388607;         //24bit整数時の振幅データMAX
                //		amp_min = -8388608;         //24bit整数時の振幅データMin

            }
#endif
#if ENABLE_KANEKO
            else if (device_type == "KANEKO") {

                string tmp_device;

                recorder = KANEKORecorder::GetInstance();

                channel_count = /*LC_CH_NUM*/4;
                sampling_rate = CLOCK_FREQ_16K;

                tmp_device = deviceName + "," + device_type;

                int ret = recorder->Initialize( tmp_device.c_str(),
                                                channel_count,
                                                sampling_rate,
                                                length * 1024);

                if (ret == -1) {
                    throw new NodeException(NULL, string("Device can not be initialized."),__FILE__, __LINE__);
                }

                hasTimestamp = true;

                amp_max = 2147483647;      //32bit整数時振の幅データMAX
                amp_min = -2.147483648e+09;//32bit整数時の振幅データMin
                //		amp_max =  8388607;         //24bit整数時の振幅データMAX
                //		amp_min = -8388608;         //24bit整数時の振幅データMin

            }
#endif
            else {
                throw new NodeException(NULL, string("Device type " + device_type + " is not supported."), __FILE__, __LINE__);
            }
            cout << "MultiAudioStreamFromMic: opening device[" << this->devices.size() << "]: " << deviceName << endl;

            Device device;
            device.deviceName = deviceName;
            device.recorder = recorder;
            this->devices.push_back(device);
        }
        inOrder = true;
    }

    ~MultiAudioStreamFromMic()
    {
        for(size_t i = 0; i < this->devices.size(); i++) {
            this->devices[i].recorder->Stop();
            delete this->devices[i].recorder;
        }
        this->devices.clear();
    }

    virtual void initialize()
    {
        if (advance > length) {
            cerr << "warning: " << name << " is set ADVANCE > LENGTH. It outputs intermittent AUDIO data." << endl;
        }

        for(size_t i = 0; i < this->devices.size(); i++) {
            outputs[this->devices[i].outputID].lookAhead = outputs[eofID].lookAhead = outputs[timeID].lookAhead = outputs[timestampID].lookAhead =
                        1
                        + max(max(max(outputs[this->devices[i].outputID].lookAhead, outputs[eofID].lookAhead), outputs[timeID].lookAhead), outputs[timestampID].lookAhead);
            //outputs[this->devices[i].outputID].lookBack += 1;

            outputs[this->devices[i].outputID].lookBack = outputs[eofID].lookBack = outputs[timeID].lookBack = outputs[timestampID].lookBack = 1
                    + max(max(max(outputs[this->devices[i].outputID].lookBack, outputs[eofID].lookBack), outputs[timeID].lookBack), outputs[timestampID].lookBack);
        }
        this->BufferedNode::initialize();
    }

    int translateOutput (const string &outputName)
    {
        size_t i;
        for(i = 0; i < this->outputNames.size(); i++) {
            if(this->outputNames[i] == outputName) {
                return i;
            }
        }

        int outputId = this->addOutput(outputName);
        int id = -1;
        char *ep = NULL;
//        cout << outputName << endl;
        if(outputName.find("AUDIO") == 0 && ((id = strtoul(outputName.c_str() + 5, &ep, 10)), ep != NULL) && *ep == '\0') {
//            cout << "AUDIO: " << id << endl;
            if(id < this->devices.size()) {
                this->devices[id].outputID = outputId;
            }
        }

        return outputId;
    }

    virtual void stop()
    {
        for(size_t i = 0; i < this->devices.size(); i++) {
            devices[i].recorder->Stop();
        }
    }

    void calculate(int output_id, int count, Buffer &out)
    {
//	cout << "Debug: count=" << count << endl;
        Buffer &eofBuffer = *(outputs[eofID].buffer);
        eofBuffer[count] = (recorder_stop_flag3 ? FalseObject : TrueObject);

        RCPtr<Time> cur_time(new Time());
        RCPtr<TimeStamp> cur_timestamp(new TimeStamp());
        (*outputs[timeID].buffer)[count] = cur_time;
        (*outputs[timestampID].buffer)[count] = cur_timestamp;

	vector<system_clock::time_point> sampleTimes;
	vector<float> avgPowers;
        for(size_t i = 0; i < this->devices.size(); i++) {
            Device &device = this->devices[i];

            RCPtr<Matrix<float> > outputp(new Matrix<float>(channel_count, length));
            Buffer &audioBuffer = *(outputs[device.outputID].buffer);
            audioBuffer[count] = outputp;
            Matrix<float>& output = *outputp;

            /* start-up device */
            if(count == 0) {
                device.recorder->Start();
                usleep(sleep_read_usec); // wait for device gets ready
            }

            /* determine frame count to be read */
            unsigned int read_offset = 0;
            unsigned int read_count  = length;
            if(count > 0) {
                read_offset = (count - 1) * advance + length;
                read_count  = advance;
            }

            /* setup buffer */
            device.buffer.resize(read_count * channel_count);

            /* fill output buffer with previously read data */
            if(count > 0 && advance < length) {
                Matrix<float>& previous = object_cast<Matrix<float> >(audioBuffer[count - 1]);
                for (int c = 0; c < length - advance; c++) {
                    for (int r = 0; r < output.nrows(); r++) {
                        output(r, c) = previous(r, c + advance);
                    }
                }
            }

            system_clock::time_point sample_time;
            while(true) {
                Recorder::BUFFER_STATE state;
//		cout << "Debug: read[" << i << "]: offset=" << read_offset << ", count=" << read_count << endl;
                if(this->hasTimestamp) {
		    state = device.recorder->ReadBuffer(read_offset, read_count, device.buffer.begin(), sample_time);
                } else {
                    state = device.recorder->ReadBuffer(read_offset, read_count, device.buffer.begin());
                }

                if(state == Recorder::OK) {
//		    cout << "Debug: read[" << i << "]: ok" << endl;
                    break;
                } else if(state == Recorder::STOPPING) {
//		    cout << "Debug: read[" << i << "]: EOF" << endl;
                    eofBuffer[count] = FalseObject;
                    return;
                } else {
//		    cout << "Debug: read[" << i << "]: wait" << endl;
                    usleep(sleep_read_usec);
                }
            }

            if(!this->hasTimestamp) {
                sample_time = std::chrono::system_clock::now();
            }
            sampleTimes.push_back(sample_time - std::chrono::microseconds((this->length - read_count) * 1000000 / this->sampling_rate));
            //cur_time->setTime(sample_time);
            //cur_timestamp->setTime(to_timeval(sample_time));

            // networkデータをfloatに変換する(2013/02/01)
            if (device_type == "NETWORK") {
                float fl1;
                for (int i = 0; i < device.buffer.size(); i++) {
                    int32_t v = (int32_t) device.buffer[i];
                    char* p = (char*) &v;
                    memcpy((char*) &fl1, p, 4);
                    device.buffer[i] = fl1 * 32767;
                }
            }

	    float psum = 0;
	    for(size_t i = 0; i < device.buffer.size(); i++){
		psum += device.buffer[i] * device.buffer[i];
	    }
	    avgPowers.push_back(sqrt(psum / device.buffer.size()) / 32768);
	    
            convertVectorToMatrix(device.buffer, output, this->length - read_count);

            bool is_clipping = false;
            for (int i = 0; i < device.buffer.size(); i++) {
                if(checkClipping(device.buffer[i])) {
                    is_clipping = true;
                    break;
                }
            }
            if(is_clipping) {
                cerr << "[" << count << "][" << getName() << "] clipping" << endl;
            }
        }

        system_clock::time_point maxSampleTime = sampleTimes[0], minSampleTime = sampleTimes[0];
        for(size_t i = 0; i < this->devices.size(); i++) {
            if(maxSampleTime < sampleTimes[i]) {
                maxSampleTime = sampleTimes[i];
            }
            if(minSampleTime > sampleTimes[i]) {
                minSampleTime = sampleTimes[i];
            }
        }
//	cout << "Debug: power: ";
//        for(size_t i = 0; i < this->devices.size(); i++) {
//	    cout << avgPowers[i];
//	    if(i == this->devices.size() - 1) {
//		cout << endl;
//	    }else{
//		cout << ", ";
//	    }
//	}
	
//	cout << "SampleTime: max=" << maxSampleTime << ", min=" << minSampleTime << ", diff=" << (maxSampleTime - minSampleTime) << endl;

	/* フレームずれ補償 */
	if(count % 100 == 0){
	    std::vector<int> current_times;
	    for(size_t i = 0; i < this->devices.size(); i++) {
		Device &device = this->devices[i];
		current_times.push_back(device.recorder->GetCurrentTime());
	    }

	    for(size_t i = 0; i < this->devices.size(); i++) {
		Device &device = this->devices[i];
		size_t frame_count = current_times[i];
		if(count == 0){
		    device.frameOffset = frame_count;
		}
		cout << "MultiAUdioStreamFromMic: frame[" << count << "], device[" << i << "]: size=" << (frame_count - device.frameOffset) << endl;
	    }

	    int min_time = current_times[0] - this->devices[0].frameOffset;
	    int max_time = current_times[0] - this->devices[0].frameOffset;
	    int max_device_id = 0;
	    for(size_t i = 0; i < this->devices.size(); i++) {
		int t = current_times[i] - this->devices[i].frameOffset;
		if(min_time > t){
		    min_time = t;
		}
		if(max_time < t){
		    max_time = t;
		    max_device_id = i;
		}
	    }

	    if(max_time - min_time >= this->frame_count_skew_tolerance){
		/* 補償が必要 */
		Recorder *recorder = this->devices[max_device_id].recorder;
		std::unique_lock<std::mutex> lk(recorder->mutex_buffer);

		recorder->cur_time -= (max_time - min_time) / 2;
	    }
	}
    }

protected:
    void ShortToFloat(const short* in, float* out, size_t n)
    {
        for (int i = 0; i < n; i++) {

            out[i] = (float) in[i];
        }
    }

    void convertVectorToMatrix(const vector<float>& in, Matrix<float>& out, int first_col)
    {
        for (int i = 0; i < out.nrows(); i++) {
            for (int j = first_col; j < out.ncols(); j++) {
                out(i, j) = (float) in[i + (j - first_col) * out.nrows()];
            }
        }
    }

    bool checkClipping(float x)
    {

        //	if (device_type == "HARKNETAPI")
        if (device_type == "NETWORK2") {
            int32_t xx;
            memcpy((char*) &xx, (char*) &x, 4);

            if (xx >= amp_max || xx <= amp_min) {
                return true;
            } else {
                return false;
            }
        } else {
            if (x >= amp_max || x <= amp_min) {
                return true;
            } else {
                return false;
            }
        }
    }


      timeval to_timeval(const std::chrono::system_clock::time_point &t) {
        std::chrono::nanoseconds diff = std::chrono::duration_cast<std::chrono::microseconds>(t.time_since_epoch());
        timeval tv;
        //drop off the fractional seconds...
        tv.tv_sec = std::chrono::duration_cast<std::chrono::seconds>(diff).count();
        //The following only works with microsecond resolution!
        tv.tv_usec = diff.count() - tv.tv_sec * 1e6;
        return tv;
      }


};
