/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <limits>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "format.h"

using namespace std;
using namespace FD;

class SaveVectorFrames;

DECLARE_NODE(SaveVectorFrames);
/*Node
 *
 * @name SaveVectorFrames
 * @category HARK:MISC
 * @description Save vector data of frames to file.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Frame Data Stream
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as input.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The file name for saving vector data.
 *
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list TEXT:RAW
 * @parameter_value TEXT
 * @parameter_description Output file format type [default: TEXT]
 *
 END*/

template <class T>
inline int printLine(const RCPtr<Vector<T> > &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);
    for (size_t i=0; i < (*v).size(); i++) {
        out << (*v)[i] << " ";
    }
    out << std::endl;
    return 0;
}

template <class T>
inline int printLine(const RCPtr<Vector<complex<T> > > &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);    
    for (size_t i=0; i < (*v).size(); i++) {
        out << (*v)[i].real() << " ";
        out << (*v)[i].imag() << " ";
    }
    out << std::endl;
}

class SaveVectorFrames : public BufferedNode {
    int inputID;
    int outputID;

    typedef enum {TEXT, RAW} OutputMode;

    OutputMode mode;
    string outputType;
    string filename;
    string formatted_filename;
    bool will_format_filename;
    ofstream fout;
    int dimension;

    public:
    SaveVectorFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT), dimension(1)
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        outputType = object_cast<String> (parameters.get("OUTPUTTYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));

        if (filename.empty()) {
            throw new NodeException(NULL, string("FILENAME is empty."), __FILE__, __LINE__);
        }

        formatted_filename.clear();
        will_format_filename = false;

        if(filename.find("{datatype", 0) != string::npos ||
           filename.find("{colsize", 0) != string::npos ||
           filename.find("{dim", 0) != string::npos) {
            will_format_filename = true;
        }

        if (outputType.compare("RAW") == 0) {
            mode = RAW;
        } else {
            mode = TEXT;
        }
    }

    ~SaveVectorFrames()
    {
        fout.close();
    }

    virtual int translateInput(const string &inputName) {
        if (inputName == "INPUT") {
            for (unsigned int i = 0; i < inputs.size(); i++) {
                if (inputs[i].name == inputName) {
                    return i;
                }
            }
        } else {
            throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
        }
    }

    string formatFilename(const char* type, int size) {
        ostringstream new_filename;

        if(will_format_filename) {
            new_filename << fmt::format(filename,
                                              fmt::arg("datatype", type),
                                              fmt::arg("colsize", size),
                                              fmt::arg("dim", dimension)); //format basename
        } else { //use original format if filename is not formatted
            new_filename << filename;
        }
        return new_filename.str();
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input_ptr;
        input_ptr = getInput(inputID, count);
        out[count] = input_ptr;

        if (typeid(*input_ptr) == typeid(Vector<float>)) {
            RCPtr<Vector<float> > input = getInput(inputID, count);
            if(!fout.is_open()) {
                formatted_filename = formatFilename("float", (*input).size());
                fout.open(formatted_filename.c_str(), ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }
            if(mode == RAW) {
                const Vector<float>& feature = object_cast<Vector<float> >(input);
                fout.write((const char*)&feature[0], sizeof(float) * feature.size());
            } else {
                printLine(input, fout);
            }
        }
        else if (typeid(*input_ptr) == typeid(Vector<complex<float> >)) {
            RCPtr<Vector<complex<float> > > input = getInput(inputID, count);
            formatted_filename = formatFilename("complex_float", (*input).size());
            if(!fout.is_open()) {
                fout.open(formatted_filename.c_str(), ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }
            if(mode == RAW) {
                const Vector<complex<float> >& feature = object_cast<Vector<complex<float> > >(input);
                fout.write((const char*)&feature[0], sizeof(float) * feature.size() * 2);
            } else {
                printLine(input, fout);
            }
        } else {
            throw new NodeException(this, "Invalid Input Type.", __FILE__, __LINE__);
        }
    }
};

